<?php

declare(strict_types=1);

namespace Fbrinker\ExtensionCheck\Extension;

use ReflectionExtension;

class ExtensionDetails
{
    private $loadedExtensions;

    /** @var ExtensionMap|null */
    private $extensionMap;

    public function __construct()
    {
        $this->loadedExtensions = array_map('strtolower', get_loaded_extensions());
    }

    /**
     * @return string[]
     */
    public function getLoadedExtensions(): array
    {
        natcasesort($this->loadedExtensions);
        return $this->loadedExtensions;
    }

    public function getLoadedExtensionsCount(): int
    {
        return count($this->loadedExtensions);
    }

    /**
     * @param string[] $excludedExtensions
     */
    public function getExtensionMap(array $excludedExtensions): ExtensionMap
    {
        if ($this->extensionMap === null) {
            $this->extensionMap = $this->buildExtensionMaps($excludedExtensions);
        }

        return $this->extensionMap;
    }

    /**
     * @param string[] $excludedExtensions
     */
    private function buildExtensionMaps(array $excludedExtensions): ExtensionMap
    {
        $extensionClasses = $extensionFunctions = $extensionConstants = $extensionDependencies = $uncheckedExtensions = [];
        foreach ($this->loadedExtensions as $loadedExtension) {
            if (in_array($loadedExtension, $excludedExtensions, true)) {
                continue;
            }

            $extension = new ReflectionExtension($loadedExtension);

            $classes = $extension->getClasses();
            if (!empty($classes)) {
                natcasesort($classes);
                foreach ($classes as $class) {
                    $extensionClasses[$class->getName()] = $loadedExtension;
                }
            }

            $functions = $extension->getFunctions();
            if (!empty($functions)) {
                natcasesort($functions);
                foreach ($functions as $function) {
                    $extensionFunctions[$function->getName()] = $loadedExtension;
                }
            }

            $constants = $extension->getConstants();
            if (!empty($constants)) {
                natcasesort($constants);
                foreach ($constants as $constant => $_) {
                    $extensionConstants[$constant] = $loadedExtension;
                }
            }

            $dependencies = $extension->getDependencies();
            if (!empty($dependencies)) {
                natcasesort($dependencies);
                foreach ($dependencies as $dependency => $status) {
                    if (in_array($dependency, $excludedExtensions, true)) {
                        continue;
                    }

                    if (!isset($extensionDependencies[$loadedExtension][strtolower($status)])) {
                        $extensionDependencies[$loadedExtension][strtolower($status)] = [];
                    }

                    $extensionDependencies[$loadedExtension][strtolower($status)][] = $dependency;
                }
            }

            // $extension->getINIEntries()
            // should we check the ini for configured extensions?

            if (empty($classes) && empty($functions)) {
                $uncheckedExtensions[] = $extension->getName();
            }
        }

        return new ExtensionMap(
            $extensionClasses,
            $extensionFunctions,
            $extensionConstants,
            $extensionDependencies,
            $uncheckedExtensions
        );
    }
}
